@@ -50,7 +50,60 @@ static char *get_cache_path() {
 	return cache_name;
 }
 
-struct string_vec compgen_cached()
+static void write_cache(const char *buffer, const char *filename)
+{
+	errno = 0;
+	FILE *fp = fopen(filename, "wb");
+	if (!fp) {
+		log_error("Failed to open cache file \"%s\": %s\n", filename, strerror(errno));
+		return;
+	}
+	size_t len = strlen(buffer);
+	errno = 0;
+	if (fwrite(buffer, 1, len, fp) != len) {
+		log_error("Error writing cache file \"%s\": %s\n", filename, strerror(errno));
+	}
+	fclose(fp);
+}
+
+static char *read_cache(const char *filename)
+{
+	errno = 0;
+	FILE *fp = fopen(filename, "rb");
+	if (!fp) {
+		log_error("Failed to open cache file \"%s\": %s\n", filename, strerror(errno));
+		return NULL;
+	}
+	if (fseek(fp, 0, SEEK_END)) {
+		log_error("Failed to seek in cache file: %s\n", strerror(errno));
+		fclose(fp);
+		return NULL;
+	}
+	size_t size;
+	{
+		long ssize = ftell(fp);
+		if (ssize < 0) {
+			log_error("Failed to determine cache file size: %s\n", strerror(errno));
+			fclose(fp);
+			return NULL;
+		}
+		size = (size_t)ssize;
+	}
+	char *cache = xmalloc(size + 1);
+	rewind(fp);
+	if (fread(cache, 1, size, fp) != size) {
+		log_error("Failed to read cache file: %s\n", strerror(errno));
+		free(cache);
+		fclose(fp);
+		return NULL;
+	}
+	fclose(fp);
+	cache[size] = '\0';
+
+	return cache;
+}
+
+char *compgen_cached()
 {
 	log_debug("Retrieving PATH.\n");
 	const char *env_path = getenv("PATH");
@@ -71,15 +124,12 @@ struct string_vec compgen_cached()
 	errno = 0;
 	if (stat(cache_path, &sb) == -1) {
 		if (errno == ENOENT) {
-			struct string_vec commands = compgen();
+			char *commands = compgen();
 			if (!mkdirp(cache_path)) {
 				free(cache_path);
 				return commands;
 			}
-			FILE *cache = fopen(cache_path, "wb");
-			string_vec_save(&commands, cache);
-			fclose(cache);
-			free(cache_path);
+			write_cache(commands, cache_path);
 			return commands;
 		}
 		free(cache_path);
@@ -103,26 +153,22 @@ struct string_vec compgen_cached()
 	}
 	free(path);
 
-	struct string_vec commands;
+	char *commands;
 	if (out_of_date) {
 		log_debug("Cache out of date, updating.\n");
 		log_indent();
 		commands = compgen();
 		log_unindent();
-		FILE *cache = fopen(cache_path, "wb");
-		string_vec_save(&commands, cache);
-		fclose(cache);
+		write_cache(commands, cache_path);
 	} else {
 		log_debug("Cache up to date, loading.\n");
-		FILE *cache = fopen(cache_path, "rb");
-		commands = string_vec_load(cache);
-		fclose(cache);
+		commands = read_cache(cache_path);
 	}
 	free(cache_path);
 	return commands;
 }
 
-struct string_vec compgen()
+char *compgen()
 {
 	log_debug("Retrieving PATH.\n");
 	const char *env_path = getenv("PATH");
@@ -167,47 +213,61 @@ struct string_vec compgen()
 	log_debug("Making unique.\n");
 	string_vec_uniq(&programs);
 
-	return programs;
+	size_t buf_len = 0;
+	for (size_t i = 0; i < programs.count; i++) {
+		buf_len += strlen(programs.buf[i].string) + 1;
+	}
+	char *buf = xmalloc(buf_len + 1);
+	size_t bytes_written = 0;
+	for (size_t i = 0; i < programs.count; i++) {
+		bytes_written += sprintf(&buf[bytes_written], "%s\n", programs.buf[i].string);
+	}
+	buf[bytes_written] = '\0';
+
+	return buf;
 }
 
-void compgen_history_sort(struct string_vec *programs, struct history *history)
+static int cmpscorep(const void *restrict a, const void *restrict b)
+{
+	struct scored_string *restrict str1 = (struct scored_string *)a;
+	struct scored_string *restrict str2 = (struct scored_string *)b;
+	return str2->history_score - str1->history_score;
+}
+
+struct string_ref_vec compgen_history_sort(struct string_ref_vec *programs, struct history *history)
 {
 	log_debug("Moving already known programs to the front.\n");
-	/*
-	 * Remove any programs in our history from the generated list, and
-	 * store which ones we found in to_add.
-	 * Removal is done without changing the count, as we're about to re-add
-	 * them at the front.
-	 */
-	struct string_vec to_add = string_vec_create();
 	for (size_t i = 0; i < history->count; i++) {
-		char **res = string_vec_find(programs, history->buf[i].name);
+		struct scored_string_ref *res = string_ref_vec_find_sorted(programs, history->buf[i].name);
 		if (res == NULL) {
+			log_debug("History entry \"%s\" not found.\n", history->buf[i].name);
 			continue;
 		}
-		free(*res);
-		*res = NULL;
-		string_vec_add(&to_add, history->buf[i].name);
-		to_add.buf[to_add.count - 1].history_score = history->buf[i].run_count;
+		res->history_score = history->buf[i].run_count;
 	}
 
-	/* Sort the vector to push the removed entries to the end. */
-	string_vec_sort(programs);
-
 	/*
-	 * Move the results down by the number of items we want to add. There's
-	 * guaranteed to be enough space to do this, as we just removed that
-	 * many items.
+	 * For compgen, we expect there to be many more commands than history
+	 * entries. For speed, we therefore create a copy of the command
+	 * vector with all of the non-zero history score items pushed to the
+	 * front. We can then call qsort() on just the first few items of the
+	 * new vector, rather than on the entire original vector.
 	 */
-	memmove(
-		&programs->buf[to_add.count],
-		programs->buf,
-		(programs->count - to_add.count) * sizeof(programs->buf[0]));
+	struct string_ref_vec vec = {
+		.count = programs->count,
+		.size = programs->size,
+		.buf = xcalloc(programs->size, sizeof(*vec.buf))
+	};
 
-	/* Add our history to the front in order. */
-	for (size_t i = 0; i < to_add.count; i++) {
-		programs->buf[i].string = xstrdup(to_add.buf[i].string);
-		programs->buf[i].history_score = to_add.buf[i].history_score;
+	size_t n_hist = 0;
+	for (ssize_t i = programs->count - 1; i >= 0; i--) {
+		if (programs->buf[i].history_score == 0) {
+			vec.buf[i + n_hist] = programs->buf[i];
+		} else {
+			vec.buf[n_hist] = programs->buf[i];
+			n_hist++;
+		}
 	}
-	string_vec_destroy(&to_add);
+	qsort(vec.buf, n_hist, sizeof(vec.buf[0]), cmpscorep);
+	return vec;
 }
