@@ -16,6 +16,27 @@
 #define MAX_INPUT_LENGTH 256
 #define MAX_PROMPT_LENGTH 256
 #define MAX_FONT_NAME_LENGTH 256
+#define MAX_FONT_FEATURES_LENGTH 128
+#define MAX_FONT_VARIATIONS_LENGTH 128
+
+struct directional {
+	int32_t top;
+	int32_t right;
+	int32_t bottom;
+	int32_t left;
+};
+
+struct text_theme {
+	struct color foreground_color;
+	struct color background_color;
+	struct directional padding;
+	uint32_t background_corner_radius;
+
+	bool foreground_specified;
+	bool background_specified;
+	bool padding_specified;
+	bool radius_specified;
+};
 
 struct entry {
 	struct image image;
@@ -34,8 +55,9 @@ struct entry {
 
 	uint32_t selection;
 	uint32_t first_result;
-	struct string_vec results;
-	struct string_vec commands;
+	char *command_buffer;
+	struct string_ref_vec results;
+	struct string_ref_vec commands;
 	struct desktop_vec apps;
 	struct history history;
 	bool use_pango;
@@ -57,7 +79,10 @@ struct entry {
 	int32_t result_spacing;
 	uint32_t font_size;
 	char font_name[MAX_FONT_NAME_LENGTH];
+	char font_features[MAX_FONT_FEATURES_LENGTH];
+	char font_variations[MAX_FONT_VARIATIONS_LENGTH];
 	char prompt_text[MAX_PROMPT_LENGTH];
+	char placeholder_text[MAX_PROMPT_LENGTH];
 	uint32_t prompt_padding;
 	uint32_t corner_radius;
 	uint32_t padding_top;
@@ -68,17 +93,22 @@ struct entry {
 	bool padding_bottom_is_percent;
 	bool padding_left_is_percent;
 	bool padding_right_is_percent;
-	int32_t selection_background_padding;
+	bool clip_to_padding;
 	uint32_t input_width;
 	uint32_t border_width;
 	uint32_t outline_width;
 	struct color foreground_color;
 	struct color background_color;
 	struct color selection_highlight_color;
-	struct color selection_foreground_color;
-	struct color selection_background_color;
 	struct color border_color;
 	struct color outline_color;
+
+	struct text_theme prompt_theme;
+	struct text_theme input_theme;
+	struct text_theme placeholder_theme;
+	struct text_theme default_result_theme;
+	struct text_theme alternate_result_theme;
+	struct text_theme selection_theme;
 };
 
 void entry_init(struct entry *entry, uint8_t *restrict buffer, uint32_t width, uint32_t height);
