@@ -26,6 +26,7 @@
 #include "shm.h"
 #include "string_vec.h"
 #include "string_vec.h"
+#include "unicode.h"
 #include "xmalloc.h"
 
 #undef MAX
@@ -33,6 +34,9 @@
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
 #define MIN(a, b) ((a) < (b) ? (a) : (b))
 
+static const char *mime_type_text_plain = "text/plain";
+static const char *mime_type_text_plain_utf8 = "text/plain;charset=utf-8";
+
 static uint32_t gettime_ms() {
 	struct timespec t;
 	clock_gettime(CLOCK_MONOTONIC, &t);
@@ -42,6 +46,44 @@ static uint32_t gettime_ms() {
 	return ms;
 }
 
+
+/* Read all of stdin into a buffer. */
+static char *read_stdin(bool normalize) {
+	const size_t block_size = BUFSIZ;
+	size_t num_blocks = 1;
+	size_t buf_size = block_size;
+
+	char *buf = xmalloc(buf_size);
+	for (size_t block = 0; ; block++) {
+		if (block == num_blocks) {
+			num_blocks *= 2;
+			buf = xrealloc(buf, num_blocks * block_size);
+		}
+		size_t bytes_read = fread(
+				&buf[block * block_size],
+				1,
+				block_size,
+				stdin);
+		if (bytes_read != block_size) {
+			if (!feof(stdin) && ferror(stdin)) {
+				log_error("Error reading stdin.\n");
+			}
+			buf[block * block_size + bytes_read] = '\0';
+			break;
+		}
+	}
+	if (normalize) {
+		if (utf8_validate(buf)) {
+			char *tmp = utf8_normalize(buf);
+			free(buf);
+			buf = tmp;
+		} else {
+			log_error("Invalid UTF-8 in stdin.\n");
+		}
+	}
+	return buf;
+}
+
 static void zwlr_layer_surface_configure(
 		void *data,
 		struct zwlr_layer_surface_v1 *zwlr_layer_surface,
@@ -96,14 +138,15 @@ static void wl_keyboard_keymap(
 {
 	struct tofi *tofi = data;
 	assert(format == WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1);
-	log_debug("Configuring keyboard.\n");
 
 	char *map_shm = mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
 	assert(map_shm != MAP_FAILED);
 
 	if (tofi->late_keyboard_init) {
+		log_debug("Delaying keyboard configuration.\n");
 		tofi->xkb_keymap_string = xstrdup(map_shm);
 	} else {
+		log_debug("Configuring keyboard.\n");
 		struct xkb_keymap *xkb_keymap = xkb_keymap_new_from_string(
 				tofi->xkb_context,
 				map_shm,
@@ -117,11 +160,10 @@ static void wl_keyboard_keymap(
 		xkb_state_unref(tofi->xkb_state);
 		tofi->xkb_keymap = xkb_keymap;
 		tofi->xkb_state = xkb_state;
+		log_debug("Keyboard configured.\n");
 	}
 	munmap(map_shm, size);
 	close(fd);
-
-	log_debug("Keyboard configured.\n");
 }
 
 static void wl_keyboard_enter(
@@ -371,6 +413,123 @@ static const struct wl_seat_listener wl_seat_listener = {
 	.name = wl_seat_name,
 };
 
+static void wl_data_offer_offer(
+		void *data,
+		struct wl_data_offer *wl_data_offer,
+		const char *mime_type)
+{
+	struct clipboard *clipboard = data;
+
+	/* Only accept plain text, and prefer utf-8. */
+	if (!strcmp(mime_type, mime_type_text_plain)) {
+		if (clipboard->mime_type != NULL) {
+			clipboard->mime_type = mime_type_text_plain;
+		}
+	} else if (!strcmp(mime_type, mime_type_text_plain_utf8)) {
+		clipboard->mime_type = mime_type_text_plain_utf8;
+	}
+}
+
+static void wl_data_offer_source_actions(
+		void *data,
+		struct wl_data_offer *wl_data_offer,
+		uint32_t source_actions)
+{
+	/* Deliberately left blank */
+}
+
+static void wl_data_offer_action(
+		void *data,
+		struct wl_data_offer *wl_data_offer,
+		uint32_t action)
+{
+	/* Deliberately left blank */
+}
+
+static const struct wl_data_offer_listener wl_data_offer_listener = {
+	.offer = wl_data_offer_offer,
+	.source_actions = wl_data_offer_source_actions,
+	.action = wl_data_offer_action
+};
+
+static void wl_data_device_data_offer(
+		void *data,
+		struct wl_data_device *wl_data_device,
+		struct wl_data_offer *wl_data_offer)
+{
+	struct clipboard *clipboard = data;
+	clipboard_reset(clipboard);
+	clipboard->wl_data_offer = wl_data_offer;
+	wl_data_offer_add_listener(
+			wl_data_offer,
+			&wl_data_offer_listener,
+			clipboard);
+}
+
+static void wl_data_device_enter(
+		void *data,
+		struct wl_data_device *wl_data_device,
+		uint32_t serial,
+		struct wl_surface *wl_surface,
+		int32_t x,
+		int32_t y,
+		struct wl_data_offer *wl_data_offer)
+{
+	/* Drag-and-drop is just ignored for now. */
+	wl_data_offer_accept(
+			wl_data_offer,
+			serial,
+			NULL);
+	wl_data_offer_set_actions(
+			wl_data_offer,
+			WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE,
+			WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE);
+}
+
+static void wl_data_device_leave(
+		void *data,
+		struct wl_data_device *wl_data_device)
+{
+	/* Deliberately left blank */
+}
+
+static void wl_data_device_motion(
+		void *data,
+		struct wl_data_device *wl_data_device,
+		uint32_t time,
+		int32_t x,
+		int32_t y)
+{
+	/* Deliberately left blank */
+}
+
+static void wl_data_device_drop(
+		void *data,
+		struct wl_data_device *wl_data_device)
+{
+	/* Deliberately left blank */
+}
+
+static void wl_data_device_selection(
+		void *data,
+		struct wl_data_device *wl_data_device,
+		struct wl_data_offer *wl_data_offer)
+{
+	struct clipboard *clipboard = data;
+	if (wl_data_offer == NULL) {
+		clipboard_reset(clipboard);
+	}
+}
+
+static const struct wl_data_device_listener wl_data_device_listener = {
+	.data_offer = wl_data_device_data_offer,
+	.enter = wl_data_device_enter,
+	.leave = wl_data_device_leave,
+	.motion = wl_data_device_motion,
+	.drop = wl_data_device_drop,
+	.selection = wl_data_device_selection
+};
+
 static void output_geometry(
 		void *data,
 		struct wl_output *wl_output,
@@ -518,12 +677,19 @@ static void registry_global(
 				&wl_shm_interface,
 				1);
 		log_debug("Bound to shm %u.\n", name);
+	} else if (!strcmp(interface, wl_data_device_manager_interface.name)) {
+		tofi->wl_data_device_manager = wl_registry_bind(
+				wl_registry,
+				name,
+				&wl_data_device_manager_interface,
+				3);
+		log_debug("Bound to data device manager  %u.\n", name);
 	} else if (!strcmp(interface, zwlr_layer_shell_v1_interface.name)) {
-		if (version < 4) {
+		if (version < 3) {
 			log_warning("Using an outdated compositor, "
 					"screen anchoring may not work.\n");
 		} else {
-			version = 4;
+			version = 3;
 		}
 		tofi->zwlr_layer_shell = wl_registry_bind(
 				wl_registry,
@@ -624,66 +790,23 @@ static const struct wl_surface_listener dummy_surface_listener = {
 };
 
 
-static void usage()
+static void usage(bool err)
 {
-	fprintf(stderr, "%s",
+	fprintf(err ? stderr : stdout, "%s",
 "Usage: tofi [options]\n"
 "\n"
+"Basic options:\n"
 "  -h, --help                           Print this message and exit.\n"
 "  -c, --config <path>                  Specify a config file.\n"
-"      --include <path>                 Include an additional config file.\n"
-"      --font <name|path>               Font to use.\n"
-"      --font-size <pt>                 Point size of text.\n"
-"      --background-color <color>       Color of the background.\n"
-"      --outline-width <px>             Width of the border outlines.\n"
-"      --outline-color <color>          Color of the border outlines.\n"
-"      --border-width <px>              Width of the border.\n"
-"      --border-color <color>           Color of the border.\n"
-"      --text-color <color>             Color of text.\n"
 "      --prompt-text <string>           Prompt text.\n"
-"      --prompt-padding <px>            Padding between prompt and input.\n"
-"      --num-results <n>                Maximum number of results to display.\n"
-"      --selection-color <color>        Color of selected result.\n"
-"      --selection-match-color <color>  Color of the matching portion of the\n"
-"                                       selected result.\n"
-"      --selection-padding <px>         Extra horizontal padding for selected\n"
-"                                       result background.\n"
-"      --selection-background <color>   Color of selected result background.\n"
-"      --result-spacing <px>            Spacing between results.\n"
-"      --min-input-width <px>           Minimum input width in horizontal mode.\n"
 "      --width <px|%>                   Width of the window.\n"
 "      --height <px|%>                  Height of the window.\n"
-"      --corner-radius <px>             Radius of window corners.\n"
 "      --output <name>                  Name of output to display window on.\n"
-"      --scale <true|false>             Follow the output's scale factor.\n"
 "      --anchor <position>              Location on screen to anchor window.\n"
-"      --exclusive-zone <-1|px|%>       Exclusive zone size, or -1 for none.\n"
-"      --margin-top <px|%>              Offset from top of screen.\n"
-"      --margin-bottom <px|%>           Offset from bottom of screen.\n"
-"      --margin-left <px|%>             Offset from left of screen.\n"
-"      --margin-right <px|%>            Offset from right of screen.\n"
-"      --padding-top <px|%>             Padding between top border and text.\n"
-"      --padding-bottom <px|%>          Padding between bottom border and text.\n"
-"      --padding-left <px|%>            Padding between left border and text.\n"
-"      --padding-right <px|%>           Padding between right border and text.\n"
-"      --hide-cursor <true|false>       Hide the cursor.\n"
 "      --horizontal <true|false>        List results horizontally.\n"
-"      --history <true|false>           Sort results by number of usages.\n"
 "      --fuzzy-match <true|false>       Use fuzzy matching for searching.\n"
-"      --require-match <true|false>     Require a match for selection.\n"
-"      --hide-input <true|false>        Hide sensitive input such as passwords.\n"
-"      --hidden-character <char>        Replacement character for hidden input.\n"
-"      --drun-launch <true|false>       Launch apps directly in drun mode.\n"
-"      --drun-print-exec <true|false>   Print a command line in drun mode.\n"
-"                                       This is now always the case,\n"
-"                                       and this option is deprecated.\n"
-"      --terminal <command>             Terminal to use for command line\n"
-"                                       programs in drun mode.\n"
-"      --hint-font <true|false>         Perform font hinting.\n"
-"      --multi-instance <true|false>    Allow multiple tofi instances at once.\n"
-"      --late-keyboard-init             (EXPERIMENTAL) Delay keyboard\n"
-"                                       initialisation until after the first\n"
-"                                       draw to screen.\n"
+"\n"
+"All options listed in \"man 5 tofi\" are also accpted in the form \"--key=value\".\n"
 	);
 }
 
@@ -698,15 +821,40 @@ const struct option long_options[] = {
 	{"corner-radius", required_argument, NULL, 0},
 	{"font", required_argument, NULL, 0},
 	{"font-size", required_argument, NULL, 0},
+	{"font-features", required_argument, NULL, 0},
+	{"font-variations", required_argument, NULL, 0},
 	{"num-results", required_argument, NULL, 0},
 	{"selection-color", required_argument, NULL, 0},
 	{"selection-match-color", required_argument, NULL, 0},
 	{"selection-padding", required_argument, NULL, 0},
 	{"selection-background", required_argument, NULL, 0},
+	{"selection-background-padding", required_argument, NULL, 0},
+	{"selection-background-corner-radius", required_argument, NULL, 0},
 	{"outline-width", required_argument, NULL, 0},
 	{"outline-color", required_argument, NULL, 0},
 	{"prompt-text", required_argument, NULL, 0},
 	{"prompt-padding", required_argument, NULL, 0},
+	{"prompt-color", required_argument, NULL, 0},
+	{"prompt-background", required_argument, NULL, 0},
+	{"prompt-background-padding", required_argument, NULL, 0},
+	{"prompt-background-corner-radius", required_argument, NULL, 0},
+	{"placeholder-text", required_argument, NULL, 0},
+	{"placeholder-color", required_argument, NULL, 0},
+	{"placeholder-background", required_argument, NULL, 0},
+	{"placeholder-background-padding", required_argument, NULL, 0},
+	{"placeholder-background-corner-radius", required_argument, NULL, 0},
+	{"input-color", required_argument, NULL, 0},
+	{"input-background", required_argument, NULL, 0},
+	{"input-background-padding", required_argument, NULL, 0},
+	{"input-background-corner-radius", required_argument, NULL, 0},
+	{"default-result-color", required_argument, NULL, 0},
+	{"default-result-background", required_argument, NULL, 0},
+	{"default-result-background-padding", required_argument, NULL, 0},
+	{"default-result-background-corner-radius", required_argument, NULL, 0},
+	{"alternate-result-color", required_argument, NULL, 0},
+	{"alternate-result-background", required_argument, NULL, 0},
+	{"alternate-result-background-padding", required_argument, NULL, 0},
+	{"alternate-result-background-corner-radius", required_argument, NULL, 0},
 	{"result-spacing", required_argument, NULL, 0},
 	{"min-input-width", required_argument, NULL, 0},
 	{"border-width", required_argument, NULL, 0},
@@ -722,9 +870,11 @@ const struct option long_options[] = {
 	{"padding-bottom", required_argument, NULL, 0},
 	{"padding-left", required_argument, NULL, 0},
 	{"padding-right", required_argument, NULL, 0},
+	{"clip-to-padding", required_argument, NULL, 0},
 	{"horizontal", required_argument, NULL, 0},
 	{"hide-cursor", required_argument, NULL, 0},
 	{"history", required_argument, NULL, 0},
+	{"history-file", required_argument, NULL, 0},
 	{"fuzzy-match", required_argument, NULL, 0},
 	{"require-match", required_argument, NULL, 0},
 	{"hide-input", required_argument, NULL, 0},
@@ -734,6 +884,7 @@ const struct option long_options[] = {
 	{"terminal", required_argument, NULL, 0},
 	{"hint-font", required_argument, NULL, 0},
 	{"multi-instance", required_argument, NULL, 0},
+	{"ascii-input", required_argument, NULL, 0},
 	{"output", required_argument, NULL, 0},
 	{"scale", required_argument, NULL, 0},
 	{"late-keyboard-init", optional_argument, NULL, 'k'},
@@ -755,14 +906,14 @@ static void parse_args(struct tofi *tofi, int argc, char *argv[])
 	int opt = getopt_long(argc, argv, short_options, long_options, &option_index);
 	while (opt != -1) {
 		if (opt == 'h') {
-			usage();
+			usage(false);
 			exit(EXIT_SUCCESS);
 		} else if (opt == 'c') {
 			config_load(tofi, optarg);
 			load_default_config = false;
 		} else if (opt == ':') {
 			log_error("Option %s requires an argument.\n", argv[optind - 1]);
-			usage();
+			usage(true);
 			exit(EXIT_FAILURE);
 		} else if (opt == '?') {
 			if (optopt) {
@@ -770,7 +921,7 @@ static void parse_args(struct tofi *tofi, int argc, char *argv[])
 			} else {
 				log_error("Unknown option %s.\n", argv[optind - 1]);
 			}
-			usage();
+			usage(true);
 			exit(EXIT_FAILURE);
 		}
 		opt = getopt_long(argc, argv, short_options, long_options, &option_index);
@@ -801,7 +952,7 @@ static void parse_args(struct tofi *tofi, int argc, char *argv[])
 
 	if (optind < argc) {
 		log_error("Unexpected non-option argument '%s'.\n", argv[optind]);
-		usage();
+		usage(true);
 		exit(EXIT_FAILURE);
 	}
 }
@@ -824,22 +975,26 @@ static bool do_submit(struct tofi *tofi)
 
 	if (entry->drun) {
 		/*
-		 * TODO: This is ugly. The list of apps needs to be sorted
-		 * alphabetically for bsearch to find the selected entry, but
-		 * we previously sorted by history count. This needs fixing.
+		 * At this point, the list of apps is history sorted rather
+		 * than alphabetically sorted, so we can't use
+		 * desktop_vec_find_sorted().
 		 */
-		desktop_vec_sort(&entry->apps);
-		struct desktop_entry *app = desktop_vec_find(&entry->apps, res);
+		struct desktop_entry *app = NULL;
+		for (size_t i = 0; i < entry->apps.count; i++) {
+			if (!strcmp(res, entry->apps.buf[i].name)) {
+				app = &entry->apps.buf[i];
+				break;
+			}
+		}
 		if (app == NULL) {
 			log_error("Couldn't find application file! This shouldn't happen.\n");
 			return false;
-		} else {
-			res = app->path;
 		}
+		char *path = app->path;
 		if (tofi->drun_launch) {
-			drun_launch(res);
+			drun_launch(path);
 		} else {
-			drun_print(res, tofi->default_terminal);
+			drun_print(path, tofi->default_terminal);
 		}
 	} else {
 		printf("%s\n", res);
@@ -848,11 +1003,79 @@ static bool do_submit(struct tofi *tofi)
 		history_add(
 				&entry->history,
 				entry->results.buf[selection].string);
-		history_save(&entry->history, entry->drun);
+		if (tofi->history_file[0] == 0) {
+			history_save_default_file(&entry->history, entry->drun);
+		} else {
+			history_save(&entry->history, tofi->history_file);
+		}
 	}
 	return true;
 }
 
+static void read_clipboard(struct tofi *tofi)
+{
+	struct entry *entry = &tofi->window.entry;
+	/* Buffer for 4 UTF-8 bytes plus a null terminator. */
+	char buffer[5];
+	memset(buffer, 0, N_ELEM(buffer));
+	errno = 0;
+	bool eof = false;
+	while (entry->input_utf32_length < N_ELEM(entry->input_utf32)) {
+		for (size_t i = 0; i < 4; i++) {
+			/*
+			 * Read input 1 byte at a time. This is slow, but easy,
+			 * and speed of pasting shouldn't really matter.
+			 */
+			int res = read(tofi->clipboard.fd, &buffer[i], 1);
+			if (res == 0) {
+				eof = true;
+				break;
+			} else if (res == -1) {
+				if (errno == EAGAIN) {
+					/*
+					 * There was no more data to be read,
+					 * but EOF hasn't been reached yet.
+					 *
+					 * This could mean more than a pipe's
+					 * capacity (64k) of data was sent, in
+					 * which case we'd potentially skip
+					 * a character, but we should hit the
+					 * input length limit long before that.
+					 */
+					input_refresh_results(tofi);
+					tofi->window.surface.redraw = true;
+					return;
+				}
+				log_error("Failed to read clipboard: %s\n", strerror(errno));
+				clipboard_finish_paste(&tofi->clipboard);
+				return;
+			}
+			uint32_t unichar = utf8_to_utf32_validate(buffer);
+			if (unichar == (uint32_t)-2) {
+				/* The current character isn't complete yet. */
+				continue;
+			} else if (unichar == (uint32_t)-1) {
+				log_error("Invalid UTF-8 character in clipboard: %s\n", buffer);
+				break;
+			} else {
+				entry->input_utf32[entry->input_utf32_length] = unichar;
+				entry->input_utf32_length++;
+				break;
+			}
+		}
+		memset(buffer, 0, N_ELEM(buffer));
+		if (eof) {
+			break;
+		}
+	}
+	entry->input_utf32[MIN(entry->input_utf32_length, N_ELEM(entry->input_utf32) - 1)] = U'\0';
+
+	clipboard_finish_paste(&tofi->clipboard);
+
+	input_refresh_results(tofi);
+	tofi->window.surface.redraw = true;
+}
+
 int main(int argc, char *argv[])
 {
 	/* Call log_debug to initialise the timers we use for perf checking. */
@@ -880,13 +1103,17 @@ int main(int argc, char *argv[])
 				.padding_bottom = 8,
 				.padding_left = 8,
 				.padding_right = 8,
+				.clip_to_padding = true,
 				.border_width = 12,
 				.outline_width = 4,
 				.background_color = {0.106f, 0.114f, 0.118f, 1.0f},
 				.foreground_color = {1.0f, 1.0f, 1.0f, 1.0f},
-				.selection_foreground_color = {0.976f, 0.149f, 0.447f, 1.0f},
 				.border_color = {0.976f, 0.149f, 0.447f, 1.0f},
 				.outline_color = {0.031f, 0.031f, 0.0f, 1.0f},
+				.placeholder_theme.foreground_color = {1.0f, 1.0f, 1.0f, 0.66f},
+				.placeholder_theme.foreground_specified = true,
+				.selection_theme.foreground_color = {0.976f, 0.149f, 0.447f, 1.0f},
+				.selection_theme.foreground_specified = true
 			}
 		},
 		.anchor =  ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP
@@ -907,6 +1134,7 @@ int main(int argc, char *argv[])
 	}
 
 	parse_args(&tofi, argc, argv);
+	log_debug("Config done\n");
 
 	if (!tofi.multiple_instance && lock_check()) {
 		log_error("Another instance of tofi is already running.\n");
@@ -920,16 +1148,20 @@ int main(int argc, char *argv[])
 	 * to Wayland.
 	 */
 
+	log_debug("Connecting to Wayland display.\n");
 	tofi.wl_display = wl_display_connect(NULL);
 	if (tofi.wl_display == NULL) {
 		log_error("Couldn't connect to Wayland display.\n");
 		exit(EXIT_FAILURE);
 	}
 	tofi.wl_registry = wl_display_get_registry(tofi.wl_display);
-	tofi.xkb_context = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
-	if (tofi.xkb_context == NULL) {
-		log_error("Couldn't create an XKB context.\n");
-		exit(EXIT_FAILURE);
+	if (!tofi.late_keyboard_init) {
+		log_debug("Creating xkb context.\n");
+		tofi.xkb_context = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
+		if (tofi.xkb_context == NULL) {
+			log_error("Couldn't create an XKB context.\n");
+			exit(EXIT_FAILURE);
+		}
 	}
 	wl_registry_add_listener(
 			tofi.wl_registry,
@@ -1098,10 +1330,18 @@ int main(int argc, char *argv[])
 	if (strstr(argv[0], "-run")) {
 		log_debug("Generating command list.\n");
 		log_indent();
-		tofi.window.entry.commands = compgen_cached();
+		tofi.window.entry.command_buffer = compgen_cached();
+		struct string_ref_vec commands = string_ref_vec_from_buffer(tofi.window.entry.command_buffer);
 		if (tofi.use_history) {
-			tofi.window.entry.history = history_load(tofi.window.entry.drun);
-			compgen_history_sort(&tofi.window.entry.commands, &tofi.window.entry.history);
+			if (tofi.history_file[0] == 0) {
+				tofi.window.entry.history = history_load_default_file(tofi.window.entry.drun);
+			} else {
+				tofi.window.entry.history = history_load(tofi.history_file);
+			}
+			tofi.window.entry.commands = compgen_history_sort(&commands, &tofi.window.entry.history);
+			string_ref_vec_destroy(&commands);
+		} else {
+			tofi.window.entry.commands = commands;
 		}
 		log_unindent();
 		log_debug("Command list generated.\n");
@@ -1111,34 +1351,39 @@ int main(int argc, char *argv[])
 		tofi.window.entry.drun = true;
 		struct desktop_vec apps = drun_generate_cached();
 		if (tofi.use_history) {
-			tofi.window.entry.history = history_load(tofi.window.entry.drun);
+			if (tofi.history_file[0] == 0) {
+				tofi.window.entry.history = history_load_default_file(tofi.window.entry.drun);
+			} else {
+				tofi.window.entry.history = history_load(tofi.history_file);
+			}
 			if (tofi.window.entry.drun) {
 				drun_history_sort(&apps, &tofi.window.entry.history);
 			}
 		}
-		struct string_vec commands = string_vec_create();
+		struct string_ref_vec commands = string_ref_vec_create();
 		for (size_t i = 0; i < apps.count; i++) {
-			string_vec_add(&commands, apps.buf[i].name);
+			string_ref_vec_add(&commands, apps.buf[i].name);
 		}
 		tofi.window.entry.commands = commands;
 		tofi.window.entry.apps = apps;
 		log_unindent();
 		log_debug("App list generated.\n");
 	} else {
-		char *line = NULL;
-		size_t n = 0;
-		tofi.window.entry.commands = string_vec_create();
-		while (getline(&line, &n, stdin) != -1) {
-			char *c = strchr(line, '\n');
-			if (c) {
-				*c = '\0';
+		log_debug("Reading stdin.\n");
+		char *buf = read_stdin(!tofi.ascii_input);
+		tofi.window.entry.command_buffer = buf;
+		tofi.window.entry.commands = string_ref_vec_from_buffer(buf);
+		if (tofi.use_history) {
+			if (tofi.history_file[0] == 0) {
+				tofi.use_history = false;
+			} else {
+				tofi.window.entry.history = history_load(tofi.history_file);
+				string_ref_vec_history_sort(&tofi.window.entry.commands, &tofi.window.entry.history);
 			}
-			string_vec_add(&tofi.window.entry.commands, line);
 		}
-		free(line);
-		tofi.use_history = false;
+		log_debug("Result list generated.\n");
 	}
-	tofi.window.entry.results = string_vec_copy(&tofi.window.entry.commands);
+	tofi.window.entry.results = string_ref_vec_copy(&tofi.window.entry.commands);
 
 	/*
 	 * Next, we create the Wayland surface, which takes on the
@@ -1195,6 +1440,19 @@ int main(int argc, char *argv[])
 			tofi.window.margin_left / tofi.window.scale);
 	wl_surface_commit(tofi.window.surface.wl_surface);
 
+	/*
+	 * Create a data device and setup a listener for data offers. This is
+	 * required for clipboard support.
+	 */
+	tofi.wl_data_device = wl_data_device_manager_get_data_device(
+			tofi.wl_data_device_manager,
+			tofi.wl_seat);
+	wl_data_device_add_listener(
+			tofi.wl_data_device,
+			&wl_data_device_listener,
+			&tofi.clipboard);
+
+
 	/*
 	 * Now that we've done all our Wayland-related setup, we do another
 	 * roundtrip. This should cause the layer surface window to be
@@ -1260,6 +1518,13 @@ int main(int argc, char *argv[])
 
 	/* If we delayed keyboard initialisation, do it now */
 	if (tofi.late_keyboard_init) {
+		log_debug("Creating xkb context.\n");
+		tofi.xkb_context = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
+		if (tofi.xkb_context == NULL) {
+			log_error("Couldn't create an XKB context.\n");
+			exit(EXIT_FAILURE);
+		}
+		log_debug("Configuring keyboard.\n");
 		struct xkb_keymap *xkb_keymap = xkb_keymap_new_from_string(
 				tofi.xkb_context,
 				tofi.xkb_keymap_string,
@@ -1273,6 +1538,7 @@ int main(int argc, char *argv[])
 		tofi.xkb_state = xkb_state;
 		free(tofi.xkb_keymap_string);
 		tofi.late_keyboard_init = false;
+		log_debug("Keyboard configured.\n");
 	}
 
 	/*
@@ -1281,8 +1547,8 @@ int main(int argc, char *argv[])
 	 * order of the various functions called here.
 	 */
 	while (!tofi.closed) {
-		struct pollfd pollfd;
-		pollfd.fd = wl_display_get_fd(tofi.wl_display);
+		struct pollfd pollfds[2] = {{0}, {0}};
+		pollfds[0].fd = wl_display_get_fd(tofi.wl_display);
 
 		/* Make sure we're ready to receive events on the main queue. */
 		while (wl_display_prepare_read(tofi.wl_display) != 0) {
@@ -1291,8 +1557,8 @@ int main(int argc, char *argv[])
 
 		/* Make sure all our requests have been sent to the server. */
 		while (wl_display_flush(tofi.wl_display) != 0) {
-			pollfd.events = POLLOUT;
-			poll(&pollfd, 1, -1);
+			pollfds[0].events = POLLOUT;
+			poll(&pollfds[0], 1, -1);
 		}
 
 		/*
@@ -1309,8 +1575,20 @@ int main(int argc, char *argv[])
 			}
 		}
 
-		pollfd.events = POLLIN | POLLPRI;
-		int res = poll(&pollfd, 1, timeout);
+		pollfds[0].events = POLLIN | POLLPRI;
+		int res;
+		if (tofi.clipboard.fd == 0) {
+			res = poll(&pollfds[0], 1, timeout);
+		} else {
+			/*
+			 * We're trying to paste from the clipboard, which is
+			 * done by reading from a pipe, so poll that file
+			 * descriptor as well.
+			 */
+			pollfds[1].fd = tofi.clipboard.fd;
+			pollfds[1].events = POLLIN | POLLPRI;
+			res = poll(pollfds, 2, timeout);
+		}
 		if (res == 0) {
 			/*
 			 * No events to process and no error - we presumably
@@ -1328,8 +1606,29 @@ int main(int argc, char *argv[])
 			/* There was an error polling the display. */
 			wl_display_cancel_read(tofi.wl_display);
 		} else {
-			/* Events to read, so put them on the queue. */
-			wl_display_read_events(tofi.wl_display);
+			if (pollfds[0].revents & (POLLIN | POLLPRI)) {
+				/* Events to read, so put them on the queue. */
+				wl_display_read_events(tofi.wl_display);
+			} else {
+				/*
+				 * No events to read - we were woken up to
+				 * handle clipboard data.
+				 */
+				wl_display_cancel_read(tofi.wl_display);
+			}
+			if (pollfds[1].revents & (POLLIN | POLLPRI)) {
+				/* Read clipboard data. */
+				if (tofi.clipboard.fd > 0) {
+					read_clipboard(&tofi);
+				}
+			}
+			if (pollfds[1].revents & POLLHUP) {
+				/*
+				 * The other end of the clipboard pipe has
+				 * closed, cleanup.
+				 */
+				clipboard_finish_paste(&tofi.clipboard);
+			}
 		}
 
 		/* Handle any events we read. */
@@ -1368,6 +1667,11 @@ int main(int argc, char *argv[])
 		wl_pointer_release(tofi.wl_pointer);
 	}
 	wl_compositor_destroy(tofi.wl_compositor);
+	if (tofi.clipboard.wl_data_offer != NULL) {
+		wl_data_offer_destroy(tofi.clipboard.wl_data_offer);
+	}
+	wl_data_device_release(tofi.wl_data_device);
+	wl_data_device_manager_destroy(tofi.wl_data_device_manager);
 	wl_seat_release(tofi.wl_seat);
 	{
 		struct output_list_element *el;
@@ -1388,8 +1692,11 @@ int main(int argc, char *argv[])
 	if (tofi.window.entry.drun) {
 		desktop_vec_destroy(&tofi.window.entry.apps);
 	}
-	string_vec_destroy(&tofi.window.entry.commands);
-	string_vec_destroy(&tofi.window.entry.results);
+	if (tofi.window.entry.command_buffer != NULL) {
+		free(tofi.window.entry.command_buffer);
+	}
+	string_ref_vec_destroy(&tofi.window.entry.commands);
+	string_ref_vec_destroy(&tofi.window.entry.results);
 	if (tofi.use_history) {
 		history_destroy(&tofi.window.entry.history);
 	}
